/** \file
 *  \brief Main form module
 */

/* Copyright (C) 2008-2010 Tomasz Ostrowski <tomasz.o.ostrowski at gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.
 */

#include <vcl.h>
#pragma hdrstop

#include "Main.h"
#include "display\DisplayOpenGL.h"
#include "display\ValueStringFormatter.h"
#include "FFT.h"
#include "FFTWindowFactory.h"
#include "Log.h"
#include "LogUnit.h"
#include "NotificationUnit.h"
#include "ScopedLock.h"
#include "Mutex.h"
#include "AboutUnit.h"
#include "MyTrackBar.h"

#include "device\MiniscopeDeviceInterface.h"
#include "device\Data.h"

#include "SettingsUnit.h"
#include "Settings.h"

#include "Recorder.h"

#include "StatisticsUnit.h"
#include "CalibrateSensitivityUnit.h"
#include "PlotToolbarUnit.h"
#include "WaitUnit.h"
#include "JsonViewerUnit.h"
#include "Ms3ViewerUnit.h"
#include "Ms3Recorder.h"
#include "Ms3FileRepairUnit.h"
#include "FilterSelector.h"
#include "DataFilterChain.h"

#include "Utilities.h"
#include "TimeCounter.h"
#include "CompileTimeChecker.h"

#include <SysUtils.hpp>
#include <ClipBrd.hpp>
#include <math.h>
#include <limits>
#include <algorithm>

//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TfrmMain *frmMain;
Settings appSettings;   ///< application settings instance
DataFilterChain filterChain;

//---------------------------------------------------------------------------
__fastcall TfrmMain::TfrmMain(TComponent* Owner)
	: TfrmCommon(Owner), fZoomChangeGrid(1.2f), device(NULL), bPaintUpdate(false),
	bFrameProgressUpdate(false),
	bJsonViewerMode(false),
	bMs3ViewerMode(false)
{
	CLog::Instance()->SetFile(ChangeFileExt(Application->ExeName, ".log").c_str());
	CLog::Instance()->LogToFile(appSettings.bLogToFile);
	CLog::Instance()->SetLevel(E_LOG_TRACE);

	display = new DisplayOpenGL(&current, pnlCenter->Handle);

	btnModeScope->Down = true;

	display->ChangeSize(pnlCenter->Width, pnlCenter->Height);

	display->SetMode(Display::MODE_OSCILLOSCOPE);
	current.SetMode(Data::MODE_OSCILLOSCOPE);
	upcoming.SetMode(Data::MODE_OSCILLOSCOPE);
	iStoredTriggerTypePos = -1;

	// load custom cursors
	Screen->Cursors[CURSOR_HAND] = LoadCursor(HInstance, "HANDCUR");
	Screen->Cursors[CURSOR_ZOOM] = LoadCursor(HInstance, "ZOOMCUR");

	mouse_tool = MOUSE_CURSOR;
	pnlCenter->Cursor = crCross;
	btnMouseCursor->Down = true;

	bMovingPlot = false;

	UpdateZoomDisplay();

	float fValX, fValY;
	display->GetCursorPhysical(Display::CURSOR1, fValX, fValY);
	UpdateCursorReadings(Display::MODE_OSCILLOSCOPE, 1, fValX, fValY);
	display->GetCursorPhysical(Display::CURSOR2, fValX, fValY);
	UpdateCursorReadings(Display::MODE_OSCILLOSCOPE, 2, fValX, fValY);

	bTriggered = false;
	bDataDivXUpdate = false;
	memset(&current, 0 , sizeof(current));

	Timer->Interval = 50;

	//STATIC_CHECK(sizeof(int) <= sizeof(char), Just_Testing);
}
//---------------------------------------------------------------------------

__fastcall TfrmMain::~TfrmMain()
{
	FFT::SetWindowType("");	// cleanup (makes CG happy)
	CLog::Instance()->Destroy();
	FFTWindowFactory::Destroy();
}

void __fastcall TfrmMain::pnlCenterResize(TObject *Sender)
{
	if (display)
		display->ChangeSize(pnlCenter->Width, pnlCenter->Height);
}
//---------------------------------------------------------------------------


void __fastcall TfrmMain::TimerTimer(TObject *Sender)
{
	Timer->Enabled = false;

	ScopedLock<Mutex> lock(mutexDeviceData);

	if (bTriggered)
	{
		ProgressBarFrameRx->Visible = true;
	}
    unsigned int iExpectedFrameSize = current.buffer_size;
	if (bDataUpdate)
	{
		if (bDataDivXUpdate)
		{
            bDataDivXUpdate = false;
			display->SetData(&current);
		}

		Data::Channel& channel = current.getChannel(0);
		if (channel.dataOsc.size() >= iExpectedFrameSize)
			current.bFrameFilled = true;
		else
			current.bFrameFilled = false;
		for (unsigned int i=0; i<current.GetChannelCount(); i++) {
			filterChain.Process(current.getChannel(i).dataOsc);
		}

		if (display->GetMode() == Display::MODE_SPECTRUM)
		{
			if (current.bFrameFilled)
			{
				if (frmStatistics->Visible)
					frmStatistics->Process(current, CurrentChannel());
				CalculateSpectrum();
			}
		}
		else if (display->GetMode() == Display::MODE_RECORDER)
		{
			if (current.bFrameFilled)
			{
				if (frmStatistics->Visible)
					frmStatistics->Process(current, CurrentChannel());
				std::vector<float>::const_iterator it;
				for (unsigned int i=0; i<current.GetChannelCount(); i++)
				{
					Data::Channel& channel = current.getChannel(i);
					float fAvg = 0;
					if (channel.dataOsc.size() > 0)
					{
						for (it = channel.dataOsc.begin(); it != channel.dataOsc.end(); ++it)
						{
							fAvg += (*it);
						}
						fAvg /= channel.dataOsc.size();
					}
					channel.dataRec.push_back(fAvg);
				}
			}
		}
		else if (display->GetMode() == Display::MODE_OSCILLOSCOPE)
		{
			if (frmStatistics->Visible)
				frmStatistics->Process(current, CurrentChannel());
		}
	}

	if (bFrameProgressUpdate || bDataUpdate)
	{
		if (iExpectedFrameSize == 0)
		{
			ProgressBarFrameRx->Position == ProgressBarFrameRx->Min;
		}
		else
		{
			if (current.bFrameFilled)
				ProgressBarFrameRx->Position = 100*upcoming.getChannel(0).dataOsc.size()/iExpectedFrameSize;
			else
				ProgressBarFrameRx->Position = 100*current.getChannel(0).dataOsc.size()/iExpectedFrameSize;
		}

		if (ProgressBarFrameRx->Position == ProgressBarFrameRx->Max)
		{
			ProgressBarFrameRx->Position = ProgressBarFrameRx->Min;
			ProgressBarFrameRx->Visible = false;
			bTriggered = false;
		}
		ProgressBarFrameRx->Update();
		bFrameProgressUpdate = false;
	}

	if (bDataUpdate || bPaintUpdate)
	{
		//TimeCounter tc("Display update");
		display->Update(Display::UPDATE_ALL);
		bPaintUpdate = false;
		bDataUpdate = false;
	}
	Timer->Enabled = true;
}
//---------------------------------------------------------------------------

void TfrmMain::SetCapabilities(const struct S_SCOPE_CAPABILITIES* const capabilities)
{
	assert(capabilities->iChannelCount > 0);
	current.SetChannelCount(capabilities->iChannelCount);
	upcoming.SetChannelCount(capabilities->iChannelCount);

	UseData(current);

	if (capabilities->iSensitivityLevelsCount > 1) {
		lblSensitivity->Visible = true;
		lblSensitivityVal->Visible = true;
		trbarSensitivity->Visible = true;
	} else {
		lblSensitivity->Visible = false;
		lblSensitivityVal->Visible = false;
		trbarSensitivity->Visible = false;
	}
	trbarSensitivity->Max = capabilities->iSensitivityLevelsCount - 1;
	trbarSensitivity->Position = trbarSensitivity->Max;

	if (capabilities->iSamplingPeriodCount > 1) {
		lblSampling->Visible = true;
		lblSamplingVal->Visible = true;
		trbarSampling->Visible = true;
	} else {
		lblSampling->Visible = false;
		lblSamplingVal->Visible = false;
		trbarSampling->Visible = false;
	}
	trbarSampling->Max = capabilities->iSamplingPeriodCount - 1;
	trbarSampling->Position = trbarSampling->Max;

	//trbarTriggerLevel->Min = pow(2, capabilities->iBitsPerSample)/(-2);
	//trbarTriggerLevel->Max = pow(2, capabilities->iBitsPerSample)/2 - 1;
	trbarTriggerLevel->Min = -100;
	trbarTriggerLevel->Max = 100;
	trbarTriggerLevel->Position = 0;

	this->cbTriggerType->Clear();
	ActionModeRecorder->Enabled = false;
	for (unsigned int i=0; i<capabilities->iTriggerTypeCount; i++) {
		AnsiString name = DeviceInterface::GetTriggerTypeName(capabilities->iTriggerTypes[i]);
		switch (capabilities->iTriggerTypes[i]) {
		case TRIGGER_MANUAL:
			cbTriggerType->Items->Add(name);
			ActionModeRecorder->Enabled = true;
			break;
		case TRIGGER_SLOPE_UP:
			cbTriggerType->Items->Add(name);
			break;
		case TRIGGER_SLOPE_DOWN:
			cbTriggerType->Items->Add(name);
			break;
		case TRIGGER_CONTINUOUS:
			cbTriggerType->Items->Add(name);
			break;
		default:
			break;
		}
	}
	if (capabilities->iTriggerTypeCount > 0) {
		cbTriggerType->ItemIndex = 0;
	}

	this->cbCouplingType->Clear();
	for (unsigned int i=0; i<capabilities->iCouplingTypeCount; i++) {
		AnsiString name = DeviceInterface::GetCouplingTypeName(capabilities->iCouplingType[i]);
		switch (capabilities->iCouplingType[i]) {
		case COUPLING_GND:
			cbCouplingType->Items->Add(name);
			break;
		case COUPLING_DC:
			cbCouplingType->Items->Add(name);
			break;
		case COUPLING_AC:
			cbCouplingType->Items->Add(name);
			break;
		default:
			break;
		}
	}
	if (capabilities->iCouplingTypeCount > 0) {
		cbCouplingType->ItemIndex = 0;
	}

	this->cbBufferSize->Clear();
	for (unsigned int i=0; i<capabilities->iBufferSizeCount; i++) {
		cbBufferSize->Items->Add(capabilities->iBufferSizes[i]);
	}
	if (capabilities->iBufferSizeCount > 0) {
		cbBufferSize->ItemIndex = 0;
	}
}
//---------------------------------------------------------------------------

void TfrmMain::CalculateSpectrum(void)
{
	unsigned int count = FFT::NearestPowerOfTwo(current.getChannel(0).dataOsc.size());
	if (count > appSettings.FFT.iSamplesLimit)
		count = appSettings.FFT.iSamplesLimit;
	FFT::SetSamplesCount(count);
	float *ReIn = new float[count];
	float *ReOut = new float[count];
	float *ImOut = new float[count];

	for (unsigned int channel_id=0; channel_id<current.GetChannelCount(); channel_id++)
	{
		std::vector<float> &samples = current.getChannel(channel_id).dataOsc;
		assert(samples.size() >= count);	// assuming same samples count for each channel
		std::vector<float> &fftv = current.getChannel(channel_id).dataFFT;
		for(unsigned int i=0;i < count;i++)
			ReIn[i]= samples[i];
		FFT::Transform(ReIn, ReOut, ImOut);
		fftv.resize(count/2);
		//LOG(E_LOG_TRACE, "Channel %d:", channel_id);
		for(unsigned int i=0;i < count/2;i++)
		{
			fftv[i] = sqrt((ReOut[i]*ReOut[i])+(ImOut[i]*ImOut[i]))/count * 2;
			//LOG(E_LOG_TRACE, "%f", fftv[i]);
		}
		if (count > 0) {
			fftv[0] /= 2.0f;
		}
	}
	delete [] ReIn;
	delete [] ReOut;
	delete [] ImOut;
}
//---------------------------------------------------------------------------


void __fastcall TfrmMain::pnlCenterMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
	if (Button == mbMiddle)
	{
		if (!frmPlotToolbar->Visible) {
			POINT P = ClientToScreen(Point(pnlCenter->Left + X, pnlCenter->Top + Y));
			frmPlotToolbar->ShowAt(P.x, P.y);
		} else
			frmPlotToolbar->Hide();
		return;
	}

	if (mouse_tool == MOUSE_CURSOR )
	{
		float fValX, fValY;
		if(Button==mbLeft)  //left button clicked
		{
			display->SetCursor1(X, Y);
			display->PosToPhysical(X, Y, fValX, fValY);
			UpdateCursorReadings(display->GetMode(), 1, fValX, fValY);
		}
		else if(Button==mbRight)
		{
			display->SetCursor2(X, Y);
			display->PosToPhysical(X, Y, fValX, fValY);
			UpdateCursorReadings(display->GetMode(), 2, fValX, fValY);
		}
	}
	else if (mouse_tool == MOUSE_OFFSET)
	{
		iPlotMoveStartX = X;
		iPlotMoveStartY = Y;
		bMovingPlot = true;
	}
	else if (mouse_tool == MOUSE_ZOOM || mouse_tool == MOUSE_ZOOM_X || mouse_tool == MOUSE_ZOOM_Y)
	{
		float fZoomX, fZoomY;
		display->GetZoom(fZoomX, fZoomY);
		float fDivX = display->GetDivX();
		float fDivY = display->GetDivY();
		if(Button==mbLeft)  //left button clicked
		{
			if (mouse_tool != MOUSE_ZOOM_Y)
			{
				if (Shift.Contains(ssShift))
				{
					fZoomX *= fZoomChangeGrid;
				}
				else
				{
					fZoomX = fZoomX * fDivX/GetNextRoundLowerValue(fDivX);
				}
			}
			if (mouse_tool != MOUSE_ZOOM_X)
			{
				if (Shift.Contains(ssShift))
				{
					fZoomY *= fZoomChangeGrid;
				}
				else
				{
					fZoomY = fZoomY * fDivY/GetNextRoundLowerValue(fDivY);
				}
			}
		}
		else if(Button==mbRight)
		{
			if (mouse_tool != MOUSE_ZOOM_Y)
			{
				if (Shift.Contains(ssShift))
				{
					fZoomX /= fZoomChangeGrid;
				}
				else
				{
					fZoomX = fZoomX * fDivX/GetNextRoundHigherValue(fDivX);
				}
			}
			if (mouse_tool != MOUSE_ZOOM_X)
			{
				if (Shift.Contains(ssShift))
				{
					fZoomY /= fZoomChangeGrid;
				}
				else
				{
					fZoomY = fZoomY * fDivY/GetNextRoundHigherValue(fDivY);
				}
			}
		}
		display->SetZoom((mouse_tool==MOUSE_ZOOM_Y)?(Display::OFFSET_LEAVE):(X),
			(mouse_tool==MOUSE_ZOOM_X)?(Display::OFFSET_LEAVE):(Y),
			fZoomX, fZoomY);
		UpdateZoomDisplay();
	}
	bPaintUpdate = true;
}
//---------------------------------------------------------------------------

float TfrmMain::GetNextRoundHigherValue(float val)
{
	assert (val > 0.0f);
	float order = log10(val);
	if (order - floor(order) > 0.9999f)	// helps with precision problem (or not?)
		order = ceil(order);
	else
		order = floor(order);
	float fNewVal = val / pow(10.0f, order);
	if (fNewVal < 1.0f)
		fNewVal = 1.0f;
	else if (fNewVal < 2.0f)
		fNewVal = 2.0f;
	else if (fNewVal < 5.0f)
		fNewVal = 5.0f;
	else
	{
		fNewVal = 1.0f;
		order += 1.0f;
	}
	fNewVal *= pow(10.0f, order);
	return fNewVal;
}

float TfrmMain::GetNextRoundLowerValue(float val)
{
	assert (val > 0.0f);
	float order = log10(val);
	if (order - floor(order) > 0.9999f)	// helps with precision problem (or not?)
		order = ceil(order);
	else
		order = floor(order);	
	float fNewVal = val / pow(10.0f, order);
	if (fNewVal > 5.0f)
		fNewVal = 5.0f;
	else if (fNewVal > 2.0f)
		fNewVal = 2.0f;
	else if (fNewVal > 1.0f)
		fNewVal = 1.0f;
	else
	{
		fNewVal = 5.0f;
		order -= 1.0f;
	}
	fNewVal *= pow(10.0f, order);
	return fNewVal;
}

void __fastcall TfrmMain::pnlCenterMouseUp(TObject *Sender, TMouseButton Button,
	  TShiftState Shift, int X, int Y)
{
	if (bMovingPlot)
	{
		if (appSettings.Graph.bSmoothTraceMoving)
		{
			display->Move(iPlotMoveStartX, iPlotMoveStartY, X, Y);
			bPaintUpdate = true;
        }
		bMovingPlot = false;
	}
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::pnlCenterMouseMove(TObject *Sender, TShiftState Shift,
	  int X, int Y)
{
	if (bMovingPlot && appSettings.Graph.bSmoothTraceMoving)
	{
		display->Move(iPlotMoveStartX, iPlotMoveStartY, X, Y);
		bPaintUpdate = true;
		iPlotMoveStartX = X;
		iPlotMoveStartY = Y;		
	}
}
//---------------------------------------------------------------------------

void TfrmMain::UpdateCursorReadings(enum Display::E_DISPLAY_MODE mode, int cursorId, float fValX, float fValY)
{
	static float fValX1 = 0.0, fValY1 = 0.0, fValX2 = 0.0, fValY2 = 0.0;

	if (mode==Display::MODE_OSCILLOSCOPE || mode==Display::MODE_RECORDER)
	{
		if (cursorId==1)
		{
			StatusBar->Panels->Items[0]->Text =
				(ValueStringFormatter::TimeToStr(fValX, 3) + " " +
				ValueStringFormatter::VoltageToStr(fValY, 3)).c_str();
			fValX1 = fValX;
			fValY1 = fValY;
		}
		else if (cursorId==2)
		{
			StatusBar->Panels->Items[1]->Text =
				(ValueStringFormatter::TimeToStr(fValX, 3) + " " +
				ValueStringFormatter::VoltageToStr(fValY, 3)).c_str();
			fValX2 = fValX;
			fValY2 = fValY;
		}
		float dx = fabs(fValX1-fValX2);
		float dy = fabs(fValY1-fValY2);
		StatusBar->Panels->Items[2]->Text =
			((std::string)"dx=" + ValueStringFormatter::TimeToStr(dx, 3) + " dy=" +
			ValueStringFormatter::VoltageToStr(dy, 3)).c_str();
		if (dx != 0)
		{
			StatusBar->Panels->Items[2]->Text = StatusBar->Panels->Items[2]->Text +
				" 1/dx=" + ValueStringFormatter::FreqToStr(1/dx, 3).c_str();
		}
	}
	else
	{
		if (cursorId==1)
		{
			StatusBar->Panels->Items[0]->Text =
				(ValueStringFormatter::FreqToStr(fValX, 3) + "  " +
				ValueStringFormatter::VoltageToStr(fValY, 3)).c_str();
			fValX1 = fValX;
			fValY1 = fValY;
		}
		else if (cursorId==2)
		{
			StatusBar->Panels->Items[1]->Text =
				(ValueStringFormatter::FreqToStr(fValX, 3) + "  " +
				ValueStringFormatter::VoltageToStr(fValY, 3)).c_str();
			fValX2 = fValX;
			fValY2 = fValY;
		}
		float dx = fabs(fValX1-fValX2);
		float dy = fabs(fValY1-fValY2);
		StatusBar->Panels->Items[2]->Text =
			((std::string)"dx=" + ValueStringFormatter::FreqToStr(dx, 3) + "  dy=" +
			ValueStringFormatter::VoltageToStr(dy, 3)).c_str();
	}
}


void __fastcall TfrmMain::ActionChangeMouseModeCursorExecute(TObject *Sender)
{
	mouse_tool = MOUSE_CURSOR;
	pnlCenter->Cursor = crCross;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ActionChangeMouseModeOffsetExecute(TObject *Sender)
{
	mouse_tool = MOUSE_OFFSET;
	pnlCenter->Cursor = (TCursor)CURSOR_HAND;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ActionChangeMouseModeZoomExecute(TObject *Sender)
{
	mouse_tool = MOUSE_ZOOM;
	pnlCenter->Cursor = (TCursor)CURSOR_ZOOM;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ActionChangeMouseModeZoomXExecute(TObject *Sender)
{
	mouse_tool = MOUSE_ZOOM_X;
	pnlCenter->Cursor = (TCursor)CURSOR_ZOOM;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ActionChangeMouseModeZoomYExecute(TObject *Sender)
{
	mouse_tool = MOUSE_ZOOM_Y;
	pnlCenter->Cursor = (TCursor)CURSOR_ZOOM;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ActionZoomClearExecute(TObject *Sender)
{
	display->SetZoom(Display::OFFSET_CLEAR, Display::OFFSET_CLEAR,
		1.0, (display->GetMode() == Display::MODE_SPECTRUM)?(2.0):(1.0));
	UpdateZoomDisplay();
	bPaintUpdate = true;	
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ActionModeScopeExecute(TObject *Sender)
{
	SetMode(Display::MODE_OSCILLOSCOPE);
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ActionModeSpectrumExecute(TObject *Sender)
{
	SetMode(Display::MODE_SPECTRUM);
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ActionModeRecorderExecute(TObject *Sender)
{
	SetMode(Display::MODE_RECORDER);
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::miAboutClick(TObject *Sender)
{
	AboutBox->ShowModal();
}
//---------------------------------------------------------------------------



void __fastcall TfrmMain::FormCreate(TObject *Sender)
{
	this->Width = appSettings.iWidth;
	this->Height = appSettings.iHeight;
	this->Top = appSettings.iPosY;
	this->Left = appSettings.iPosX;
	if (appSettings.bAlwaysOnTop)
		this->FormStyle = fsStayOnTop;
	else
		this->FormStyle = fsNormal;
	if (appSettings.bWindowMaximized)
		this->WindowState = wsMaximized;
	edInterval->Text = AnsiString::FloatToStrF(appSettings.fRecorderInterval, AnsiString::sffGeneral, 1, 3);
	FFT::SetWindowType(appSettings.FFT.asWindowName);
	((TMyTrackBar*)trbarSensitivity)->OnMouseUp = TrackBarMouseUp;
	((TMyTrackBar*)trbarSampling)->OnMouseUp = TrackBarMouseUp;
	((TMyTrackBar*)trbarTriggerLevel)->OnMouseUp = TrackBarMouseUp;
	((TMyTrackBar*)trbarTriggerPosition)->OnMouseUp = TrackBarMouseUp;


	TMenuItem* mi = new TMenuItem(miLanguage);
	mi->Caption = "en";
	mi->OnClick = miSelectLangClick;
	mi->Checked = (mi->Caption == appSettings.asLanguage);
	mi->RadioItem = true;
	miLanguage->Add(mi);

	Classes::TStrings* list = new Classes::TStringList();
	DefaultInstance->GetListOfLanguages("default", list);
	for (int i=0; i<list->Count; i++)
	{
		TMenuItem* mi = new TMenuItem(miLanguage);
		mi->Caption = list->Strings[i];
		mi->OnClick = miSelectLangClick;
		mi->Checked = (mi->Caption == appSettings.asLanguage);
		mi->RadioItem = true;
		miLanguage->Add(mi);
	}
	delete list;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::miSelectLangClick(TObject *Sender)
{
	TMenuItem *clicked;
	clicked = dynamic_cast<TMenuItem*>(Sender);
	if (!clicked)
		return;
	// deselect all menu items
	for (int i = miLanguage->Count - 1; i>=0; --i)
	{
		miLanguage->Items[i]->Checked = false;
	}
	clicked->Checked = true;
	appSettings.asLanguage = clicked->Caption;

#if 0
	// we could try to re-translate, but complete re-translation
	// would require keeping track of i.e. trigger names added to combobox and
	// recreating them each time language is switched
	//UseLanguage(appSettings.asLanguage);
	//RetranslateComponent(this);
#else
	// I think this message should not be translated
	ShowMessage("New language will be used after restart.");
#endif
}

void __fastcall TfrmMain::miDeviceSettingsClick(TObject *Sender)
{
	device->ShowSettings();
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::miViewLogClick(TObject *Sender)
{
	frmLog->Visible = miViewLog->Checked;
	if (frmLog->Visible)
		frmLog->BringToFront();
}
//---------------------------------------------------------------------------

void TfrmMain::CallbackCloseLogFn(void)
{
	miViewLog->Checked = false;
}

void TfrmMain::CallbackDataFn(short* buf, int size)
{
	/** \note DO NOT use VCL code inside callbacks called by device dll */
	ScopedLock<Mutex> lock(mutexDeviceData);
	size /= sizeof(*buf);
	if (bJsonViewerMode || bMs3ViewerMode)
		return;
#if 0
	AnsiString log = "";
	int log_cnt = size;
	if (log_cnt > 20)
		log_cnt = 20;
	for (int i=0; i<log_cnt; i++) {
		log = log + " " + buf[i];
	}
	LOG(E_LOG_TRACE, "Data (%d): %s", size, log.c_str());
	//LOG(E_LOG_TRACE, "Data: %d", size);
	//LOG(E_LOG_TRACE, "DataFn: %d samples, total: %d samples", size, dataOscTmp.size());
#endif

	short* ptr = buf;
	// push received samples to "upcoming" temporary vector
	for (unsigned int channel_id=0; channel_id<upcoming.GetChannelCount(); channel_id++)
	{
		Data::Channel& channel = upcoming.getChannel(channel_id);
		float coeff = upcoming.getChannel(channel_id).fDivY * upcoming.getChannel(channel_id).calibration.gain;
		float offset = upcoming.getChannel(channel_id).calibration.offset;
		for (unsigned int i=0; i<size/upcoming.GetChannelCount(); i++)
		{
			channel.dataOsc.push_back(((*ptr)-offset) * coeff);
			ptr++;
		}
	}

	if (current.bFrameFilled)
	{
		if (upcoming.getChannel(0).dataOsc.size() >= upcoming.buffer_size)
		{
			current = upcoming;
			for (unsigned int i=0; i<upcoming.GetChannelCount(); i++)
			{
				upcoming.getChannel(i).dataOsc.clear();
				upcoming.getChannel(i).dataOsc.reserve(102400);
			}
			bDataUpdate = true;
			if (Ms3Recorder::Instance().GetMode() == Ms3Recorder::MODE_WRITE)
			{
				Ms3Recorder::Instance().StoreFrame(current);
				if (Ms3Recorder::Instance().GetFileSize() > (unsigned int)appSettings.Ms3Recording.iSizeLimit*1024)
				{
					AnsiString msg = _("MS3 file size reached (check settings), recording stopped");
					TfrmNotification::Add(TfrmNotification::WARNING, msg.c_str());
					ActionMs3Record->Execute();
				}
			}
		}
		else
		{
			bFrameProgressUpdate = true;
		}
	}
	else
	{
		bool bStoreFrame = false;
		current.fOscDivX = upcoming.fOscDivX;
		current.timestamp = upcoming.timestamp;
		current.trigger_source = upcoming.trigger_source;
		current.trigger_type = upcoming.trigger_type;
		current.buffer_size = upcoming.buffer_size;
		for (unsigned int i=0; i<upcoming.GetChannelCount(); i++)
		{
			Data::Channel& channelDst = current.getChannel(i);
			Data::Channel& channelSrc = upcoming.getChannel(i);
			int end = channelSrc.dataOsc.size();
			for (int i=channelDst.dataOsc.size(); i<end; i++)
			{
				channelDst.dataOsc.push_back(channelSrc.dataOsc[i]);
			}
			if (channelSrc.dataOsc.size() >= upcoming.buffer_size)
			{
				channelSrc.dataOsc.clear();
				channelSrc.dataOsc.reserve(102400);
				bStoreFrame = true;
			}
		}
		if (bStoreFrame)
		{
			if (Ms3Recorder::Instance().GetMode() == Ms3Recorder::MODE_WRITE)
				Ms3Recorder::Instance().StoreFrame(current);
        }
		bDataUpdate = true;
	}
}

void TfrmMain::CallbackTriggerFn(void)
{
	ScopedLock<Mutex> lock(mutexDeviceData);
	if (bJsonViewerMode || bMs3ViewerMode)
		return;
	ftime(&upcoming.timestamp);

#if 0
	LOG(E_LOG_TRACE, "Trigger");
#endif

	/** \note DO NOT use VCL code inside callbacks called by device dll */

	if (current.fOscDivX != upcoming.fOscDivX ||
		current.trigger_type != upcoming.trigger_type ||
/** \todo multichannel: disable channels */
//		memcmp(&current.channels[0], &upcoming.channels[0], current.channels.size() * sizeof(current.channels[0])),
		current.buffer_size != upcoming.buffer_size)
	{
		// clear display if any acquisition settings changed
		for (unsigned int i=0; i<current.GetChannelCount(); i++)
		{
			Data::Channel& channel = current.getChannel(i);
			channel.dataOsc.clear();
			channel.dataOsc.reserve(102400);
		}
		bDataUpdate = true;
		if (current.fOscDivX != upcoming.fOscDivX)
			bDataDivXUpdate = true;
		current = upcoming;
	}
	for (unsigned int i=0; i<upcoming.GetChannelCount(); i++)
	{
		Data::Channel& channel = upcoming.getChannel(i);
		channel.dataOsc.clear();
		channel.dataOsc.reserve(102400);
	}

	bTriggered = true;
}

void TfrmMain::CallbackConnectFn(int state, char *szMsgText)
{
	tmrUpdateConnState->Enabled = true;
}

void TfrmMain::UpdateConnectionState(void)
{
	struct ConnectionInfo conn = device->GetConnectionInfo();
	pnlOscControl->Enabled = false;
	switch (conn.state)
	{
	case DEVICE_DISCONNECTED:
		break;
	case DEVICE_CONNECTING:
		break;
	case DEVICE_CONNECTED:
		pnlOscControl->Enabled = true;
		break;
	case DEVICE_DISCONNECTING:
		break;
	default:
		break;
	}
    StatusBar->Panels->Items[3]->Text = conn.msg;
}

void __fastcall TfrmMain::ActionAcquireStartExecute(TObject *Sender)
{
	device->AcquireStart();
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ActionAcquireStopExecute(TObject *Sender)
{
	device->AcquireStop();
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::tmrUpdateConnStateTimer(TObject *Sender)
{
	tmrUpdateConnState->Enabled = false;
	UpdateConnectionState();
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::cbTriggerChange(TObject *Sender)
{
	UpdateTriggerType(true);
}
//---------------------------------------------------------------------------

void TfrmMain::UpdateTriggerType(bool bUpdateDevice)
{
	enum E_TRIGGER_TYPE type = device->capabilities->iTriggerTypes[cbTriggerType->ItemIndex];
	int channel = cbTriggerSource->ItemIndex;
	if (bUpdateDevice)
		device->SetTriggerType(channel, cbTriggerType->ItemIndex);
	btnManualTrigger->Visible = false;
	trbarTriggerLevel->Visible = false;
	lblTriggerLevel->Visible = false;
	lblTriggerLevelVal->Visible = false;
	trbarTriggerPosition->Visible = false;
	lblTriggerPosition->Visible = false;
	lblTriggerPositionVal->Visible = false;
	lblTriggerSource->Visible = false;
	cbTriggerSource->Visible = false;
	btnStart->Visible = false;
	btnStop->Visible = false;
	display->SetTriggerChannel(channel);
	if ((type == TRIGGER_SLOPE_UP || type == TRIGGER_SLOPE_DOWN) && device->hasTriggerPositionControl())
	{
		display->SetVisibility(Display::TRIGGER_POS_MARKER, appSettings.GraphVisibility.bTriggerPosMarker);
	}
	else
	{
		display->SetVisibility(Display::TRIGGER_POS_MARKER, false);
	}
	if (type == TRIGGER_SLOPE_UP || type == TRIGGER_SLOPE_DOWN)
	{
		display->SetVisibility(Display::TRIGGER_VAL_MARKER, appSettings.GraphVisibility.bTriggerValMarker);
	}
	else
	{
		display->SetVisibility(Display::TRIGGER_VAL_MARKER, false);
	}
	bPaintUpdate = true;
	switch (type) {
	case TRIGGER_MANUAL:
		btnManualTrigger->Visible = true;
		break;
	case TRIGGER_SLOPE_UP:
	case TRIGGER_SLOPE_DOWN:
		trbarTriggerLevel->Visible = true;
		lblTriggerLevel->Visible = true;
		lblTriggerLevelVal->Visible = true;
		if (device->hasTriggerPositionControl())
		{
			trbarTriggerPosition->Visible = true;
			lblTriggerPosition->Visible = true;
			lblTriggerPositionVal->Visible = true;
		}
		lblTriggerSource->Visible = true;
		cbTriggerSource->Visible = true;
		break;
	case TRIGGER_CONTINUOUS:
		btnStart->Visible = true;
		btnStop->Visible = true;
		break;
	default:
		break;
	}
	upcoming.trigger_type = device->capabilities->iTriggerTypes[cbTriggerType->ItemIndex];
}

void TfrmMain::UpdateCouplingType(void)
{
    int channel = tabOscControlTop->TabIndex;
	device->SetCouplingType(channel, cbCouplingType->ItemIndex);
	upcoming.getChannel(channel).iCouplingType = device->capabilities->iCouplingType[cbCouplingType->ItemIndex];
}

void __fastcall TfrmMain::btnManualTriggerClick(TObject *Sender)
{
	int status = device->ManualTrigger();
	if (status)
		LOG(E_LOG_TRACE, "ManualTrigger() failed with status %d", status);
	else
		ProgressBarFrameRx->Visible = true;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::cbBufferSizeChange(TObject *Sender)
{
	UpdateBufferSize();
}
//---------------------------------------------------------------------------

void TfrmMain::UpdateBufferSize(void)
{
	if (!cbBufferSize->Items->Count)
		return;
	device->SetBufferSize(cbBufferSize->ItemIndex);
	upcoming.buffer_size = device->capabilities->iBufferSizes[cbBufferSize->ItemIndex];
}

void TfrmMain::UpdateSensitivity(bool bUpdateDevice)
{
    int channel = tabOscControlTop->TabIndex;
	int range = trbarSensitivity->Position;
	if (bUpdateDevice)
		device->SetSensitivity(channel, range);
	// update range caption, lblSensitivityVal
	float fDivVal = (float)pow(2, device->capabilities->iBitsPerSample) / (float)10 *
		device->capabilities->fSensitivityLevels[range] *
		calibration[channel][range].gain;
	lblSensitivityVal->Caption =
		ValueStringFormatter::VoltageToStr(fDivVal, 1).c_str();
	lblSensitivityVal->Caption = lblSensitivityVal->Caption + "/div";
	upcoming.getChannel(channel).fDivY = device->capabilities->fSensitivityLevels[range];
	upcoming.getChannel(channel).calibration = calibration[channel][range];

	if (display)
	{
		display->SetRange(channel, fDivVal * 5.0);
		//display->SetDiv(device->capabilities->fSamplingPeriod[trbarSampling->Position]);
		bDataUpdate = true;
	}
	frmCalibrateSensitivity->SetCurrentCalibration(
		calibration[channel][range].offset,
		calibration[channel][range].gain);
}

void TfrmMain::UpdateSampling(bool bUpdateDevice)
{
	if (bUpdateDevice)
		device->SetSampling(trbarSampling->Position);
	float fSamplVal = (float)1.0 / device->capabilities->fSamplingPeriod[trbarSampling->Position];
	lblSamplingVal->Caption =
		ValueStringFormatter::FreqToStr(fSamplVal, 1).c_str();
	upcoming.fOscDivX = device->capabilities->fSamplingPeriod[trbarSampling->Position];
}

void TfrmMain::UpdateTriggerLevel(void)
{
	int val = trbarTriggerLevel->Position;
	device->SetTriggerLevel(val);
	lblTriggerLevelVal->Caption = IntToStr(val) + "%";
	display->SetTriggerValue(val);
}

void TfrmMain::UpdateTriggerPosition(void)
{
	int pos = trbarTriggerPosition->Position;
	device->SetTriggerPosition(pos);
	lblTriggerPositionVal->Caption = IntToStr(pos) + "%";
	display->SetTriggerPosition(pos);
}

void __fastcall TfrmMain::cbCouplingTypeChange(TObject *Sender)
{
	UpdateCouplingType();
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::PaintBoxCenterPaint(TObject *Sender)
{
	/** \note PaintBoxCenter purpose is to fire OnPaint event
		to request graph repaint (TPanel has no OnPaint event but
		TPaintBox has no Handle)
    */
	bPaintUpdate = true;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::PaintBoxCenterMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
	pnlCenterMouseDown(Sender, Button, Shift, X, Y);
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::PaintBoxCenterMouseMove(TObject *Sender,
	  TShiftState Shift, int X, int Y)
{
	pnlCenterMouseMove(Sender, Shift, X, Y);
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::PaintBoxCenterMouseUp(TObject *Sender,
	  TMouseButton Button, TShiftState Shift, int X, int Y)
{
	pnlCenterMouseUp(Sender, Button, Shift, X, Y);
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::FormCloseQuery(TObject *Sender, bool &CanClose)
{
	Timer->Enabled = false;
	
	int status = Ms3Recorder::Instance().SetMode(Ms3Recorder::MODE_IDLE);
	ReportError(status, ((AnsiString)(_("Failed to close MS3 recorder/player."))).c_str());

	delete display;
	if (device)
	{
		device->SaveSettings(&device->settings);
		device->Disconnect();
		delete device;
		device = NULL;
	}

	filterChain.Release();

	AnsiString asConfigFile = ChangeFileExt( Application->ExeName, ".ini" );
	appSettings.bWindowMaximized = (this->WindowState == wsMaximized);
	if (!appSettings.bWindowMaximized)
	{
		// these values are meaningless is wnd is maximized
		appSettings.iWidth = this->Width;
		appSettings.iHeight = this->Height;
		appSettings.iPosY = this->Top;
		appSettings.iPosX = this->Left;
	}

	appSettings.Statistics.iPosX = frmStatistics->Left;
	appSettings.Statistics.iPosY = frmStatistics->Top;
	appSettings.Statistics.bVisible = frmStatistics->Visible;

	appSettings.Write(asConfigFile);

	CanClose = true;
}
//---------------------------------------------------------------------------

void TfrmMain::LoadDevice(AnsiString dllname)
{
    // get rid of artifacts first
	frmMain->Repaint();
	display->Update(Display::UPDATE_ALL);
	frmWait->Show(_("Loading input device dll, please wait..."));

//	ToolBarMode->Enabled = false;
	if (device)
	{
		device->SaveSettings(&device->settings);
		device->Disconnect();
		delete device;
		device = NULL;
	}
	iStoredTriggerTypePos = -1;
	device = new DeviceInterface(dllname);
	if (device->Load())
	{
		ShowMessage(_("Failed to load device dll."));
		delete device;
		device = NULL;
		frmWait->Close();
		return;
	}
	device->callbackData = CallbackDataFn;
	device->callbackConnect = CallbackConnectFn;
	device->callbackTrigger = CallbackTriggerFn;

	current.fOscDivX =device->capabilities->fSamplingPeriod[0];
	display->SetData(&current);
	SetCapabilities(device->capabilities);

	int current_channel = tabOscControlTop->TabIndex;
	if (device->GetSettings(&device->settings) == 0)
	{
		trbarSensitivity->Position = device->settings.pSensitivity[current_channel];
		trbarSampling->Position = device->settings.iSamplingPeriod;
		trbarTriggerLevel->Position = device->settings.iTriggerLevel;
		trbarTriggerPosition->Position = device->settings.iTriggerPosition;
		for (unsigned int i=0; i<device->capabilities->iTriggerTypeCount; i++)
		{
			if (device->capabilities->iTriggerTypes[i] == device->settings.eTriggerType)
			{
				cbTriggerType->ItemIndex = i;
				break;
			}
		}
		for (unsigned int i=0; i<device->capabilities->iCouplingTypeCount; i++)
		{
			if (device->capabilities->iCouplingType[i] == device->settings.eCouplingType[current_channel])
			{
				cbCouplingType->ItemIndex = i;
				break;
			}
		}
		cbBufferSize->ItemIndex = cbBufferSize->Items->IndexOf(device->settings.iBufferSize);
		cbTriggerSource->ItemIndex = device->settings.iTriggerSrcChannel;
	}
	else
	{
		cbTriggerSource->ItemIndex = 0;
	}

	ActionCalibrateSensitivity->Enabled = device->hasCalibration();
	calibration.clear();
	if (device->hasCalibration())
	{
		for (unsigned int i=0; i<device->capabilities->iChannelCount; i++)
		{
			calibration.push_back ( std::vector<struct DEV_CALIBRATION>() );
			for (unsigned int j=0; j<device->capabilities->iSensitivityLevelsCount; j++)
			{
				struct DEV_CALIBRATION cal = {0, 1.0};
				device->GetCalibration(i, j, &cal.offset, &cal.gain);
				calibration[i].push_back(cal);
			}
		}
		//ApplyCalibration();
	} else {
		for (unsigned int i=0; i<device->capabilities->iChannelCount; i++)
		{
            calibration.push_back ( std::vector<struct DEV_CALIBRATION>() );
			for (unsigned int j=0; j<device->capabilities->iSensitivityLevelsCount; j++)
			{
				struct DEV_CALIBRATION cal = {0, 1.0};
				calibration[i].push_back(cal);
			}
		}
	}

	frmWait->Show(_("Connecting to input device..."));

	if (device->Connect())
	{
		frmWait->Close();
		// update only GUI state if required
		UpdateTriggerType(false);
		UpdateSensitivity(false);
		UpdateSampling(false);
		ShowMessage(_("Failed connecting to device."));
	} else
	{
		UpdateTriggerType();
		UpdateBufferSize();
		for (unsigned int i=0; i<device->capabilities->iChannelCount; i++)
		{
			device->SetSensitivity(i, device->settings.pSensitivity[i]);
			float fDivVal = (float)pow(2, device->capabilities->iBitsPerSample) / (float)10 *
				device->capabilities->fSensitivityLevels[device->settings.pSensitivity[i]] *
				calibration[i][device->settings.pSensitivity[i]].gain;
			display->SetRange(i, fDivVal * 5.0);
			upcoming.getChannel(i).fDivY = device->capabilities->fSensitivityLevels[device->settings.pSensitivity[i]];
			upcoming.getChannel(i).calibration = calibration[i][device->settings.pSensitivity[i]];
			upcoming.getChannel(i).iCouplingType = device->settings.eCouplingType[i];
			for (unsigned int j=0; j<device->capabilities->iCouplingTypeCount; j++)
			{
				if (device->capabilities->iCouplingType[j] == device->settings.eCouplingType[i])
				{
					device->SetCouplingType(i, j);
					break;
				}
			}
		}
		UpdateSensitivity();
		UpdateCouplingType();
		UpdateSampling();
		UpdateTriggerLevel();
		UpdateTriggerPosition();
//		ToolBarMode->Enabled = true;
		frmWait->Close();
	}
	current = upcoming;
	bDataDivXUpdate = true;
}

std::auto_ptr<Graphics::TBitmap> TfrmMain::CreatePlotBitmap(void)
{
	bool bStatState = frmStatistics->Visible;
	if (bStatState)
	{
        // hide statistics window
		frmStatistics->Visible = false;
		display->Update(Display::UPDATE_ALL);
	}
	frmStatistics->Update();
	std::auto_ptr<Graphics::TBitmap> bitmap = display->CreateBitmap();
	if (bStatState)
	{
		frmStatistics->Visible = bStatState;
		this->SetFocus();
	}
	return bitmap;
}

void __fastcall TfrmMain::ActionCopyPlotClipbrdExecute(TObject *Sender)
{
	// make sure display is repainted
	display->Update(Display::UPDATE_ALL);
	std::auto_ptr<Graphics::TBitmap> bitmap = CreatePlotBitmap();
	TClipboard *pCB = Clipboard();
	pCB->Assign(&(*bitmap));
	(void)pCB;	// avoid unused variable warning
}
//---------------------------------------------------------------------------

void TfrmMain::UpdateZoomDisplay(void)
{
	float fZoomX, fZoomY;
	display->GetZoom(fZoomX, fZoomY);
	lblZoomX->Caption = (AnsiString)"X: " + (ValueStringFormatter::ValToStr(fZoomX, 3) + "x").c_str();
	lblZoomY->Caption = (AnsiString)"Y: " + (ValueStringFormatter::ValToStr(fZoomY, 3) + "x").c_str();

	if (display->GetMode() == Display::MODE_SPECTRUM)
		lblXscale->Caption = ValueStringFormatter::FreqToStr(display->GetDivX(), 2).c_str();
	else
		lblXscale->Caption = ValueStringFormatter::TimeToStr(display->GetDivX(), 2).c_str();
	lblYscale->Caption = ValueStringFormatter::VoltageToStr(display->GetDivY(), 2).c_str();
	lblXscale->Caption = lblXscale->Caption + "/div";
	lblYscale->Caption = lblYscale->Caption + "/div";
}


void __fastcall TfrmMain::TrackbarKeyUp(TObject *Sender, WORD &Key,
	  TShiftState Shift)
{
	UpdateTrackBar(Sender);
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::TrackBarMouseUp(TObject *Sender,
	TMouseButton Button, TShiftState Shift, int X, int Y)
{
	UpdateTrackBar(Sender);
}

void TfrmMain::UpdateTrackBar(TObject *Sender)
{
	if (Sender == trbarSensitivity)
	{
		UpdateSensitivity();
	}
	else if (Sender == trbarSampling)
	{
		UpdateSampling();
	}
	else if (Sender == trbarTriggerLevel)
	{
		UpdateTriggerLevel();
	}
	else if (Sender == trbarTriggerPosition)
	{
    	UpdateTriggerPosition();
	}
}

void __fastcall TfrmMain::miStatisticsClick(TObject *Sender)
{
	frmStatistics->Show();
	ScopedLock<Mutex> lock(mutexDeviceData);

	unsigned int iExpectedFrameSize = device->capabilities->iBufferSizes[cbBufferSize->ItemIndex];
	if (current.getChannel(0).dataOsc.size() == iExpectedFrameSize)
	{
		frmStatistics->Process(current, CurrentChannel());
	}
}
//---------------------------------------------------------------------------

void TfrmMain::FirstRunDialog(void)
{
	AnsiString msg = _("No input device dll is selected.\n"
		"It seems like application is running for ther first time\n"
		"or configuration has been removed.\n"
		"Selecting device library is REQUIRED.");
	AnsiString title = _("No input device plugin selected");
	MessageBox(this->Handle, msg.c_str(), title.c_str(), MB_OK | MB_ICONINFORMATION);
	    ActionShowCommonSettingsExecute(NULL);
}

void __fastcall TfrmMain::ActionShowCommonSettingsExecute(TObject *Sender)
{
	AnsiString asPrevDevice = appSettings.asInputDeviceDllName;
	frmSettings->appSettings = &appSettings;
	frmSettings->ShowModal();
	if (appSettings.bAlwaysOnTop)
		this->FormStyle = fsStayOnTop;
	else
		this->FormStyle = fsNormal;
	if (appSettings.asInputDeviceDllName != asPrevDevice)
		LoadDevice(asDevDllDir + "\\" + appSettings.asInputDeviceDllName);
	Timer->Interval = 1000/appSettings.iDisplayRefreshFreq;
	
	display->SetChannelColors(appSettings.GraphColors.traces);
	tabOscControlTop->Invalidate();		
	frmPlotToolbar->SetTransparecy(appSettings.Graph.ToolboxTransparency);
	FFT::SetWindowType(appSettings.FFT.asWindowName);
	CLog::Instance()->LogToFile(appSettings.bLogToFile);
	bPaintUpdate = true;
}
//---------------------------------------------------------------------------

void TfrmMain::SetMode(enum Display::E_DISPLAY_MODE mode)
{
	int oldmode = -1;

    // restore trigger type to state before switching to recorder mode
	if (mode != oldmode && oldmode == Display::MODE_RECORDER)
	{
		if (cbTriggerType->Items->Count > iStoredTriggerTypePos && iStoredTriggerTypePos != -1)
		{
			cbTriggerType->ItemIndex = iStoredTriggerTypePos;		
			UpdateTriggerType();
		}
	}

	bDataUpdate = true;
	if (mode == Display::MODE_RECORDER)
	{
		iStoredTriggerTypePos = cbTriggerType->ItemIndex;
		for (unsigned int i=0; i<device->capabilities->iTriggerTypeCount; i++)
		{
			if (device->capabilities->iTriggerTypes[i] == TRIGGER_MANUAL)
			{
				cbTriggerType->ItemIndex = i;
				break;
			}
		}
		UpdateTriggerType();
	}
	display->SetMode(mode);
	UpdateZoomDisplay();

	if (mode == Display::MODE_OSCILLOSCOPE)
		current.SetMode(Data::MODE_OSCILLOSCOPE);
	else if (mode == Display::MODE_RECORDER)
		current.SetMode(Data::MODE_RECORDER);
	else if (mode == Display::MODE_SPECTRUM)
		current.SetMode(Data::MODE_SPECTRUM);
	else {
		assert(0);
	}
	upcoming.SetMode(current.GetMode());

	if (mode != Display::MODE_RECORDER)
	{
		lblTriggerType->Visible = true;
		cbTriggerType->Visible = true;
		if (device->capabilities->iTriggerTypes[cbTriggerType->ItemIndex] == TRIGGER_MANUAL)
			btnManualTrigger->Visible = true;
		btnRecorderStart->Visible = false;
		btnRecorderStop->Visible = false;
		btnRecorderClear->Visible = false;
		lblInterval->Visible = false;
		edInterval->Visible = false;
		lblSeconds->Visible = false;

		tmrRecorder->Enabled = false;
		btnRecorderStart->Enabled = true;
		btnRecorderStop->Enabled = false;
	}
	else
	{
        lblTriggerType->Visible = false;
		cbTriggerType->Visible = false;
		btnManualTrigger->Visible = false;
		btnRecorderStart->Visible = true;
		btnRecorderStop->Visible = true;
		btnRecorderClear->Visible = true;
		lblInterval->Visible = true;
		edInterval->Visible = true;
		lblSeconds->Visible = true;
	}	
}



void __fastcall TfrmMain::btnRecorderStartClick(TObject *Sender)
{
	float interval = StrToFloatDef(edInterval->Text, -1.0);
	if (interval < Settings::fRecorderMinInterval || interval > Settings::fRecorderMaxInterval)
	{
		AnsiString msg = _("Recorder interval outside allowed range!");
		MessageBox(this->Handle, msg.c_str(), this->Caption.c_str(),
			MB_ICONEXCLAMATION);
		//edInterval->Text = AnsiString::FloatToStrF(appSettings.fRecorderInterval, AnsiString::sffGeneral, 1, 3);
		return;
	}
	appSettings.fRecorderInterval = interval;
	current.fRecorderInterval = interval;
	display->SetData(&current);

	tmrRecorder->Interval = interval * 1000.0;


	tmrRecorder->Enabled = true;
	btnRecorderStart->Enabled = false;
	edInterval->Enabled = false;
	btnRecorderStop->Enabled = true;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::btnRecorderStopClick(TObject *Sender)
{
	tmrRecorder->Enabled = false;
	btnRecorderStart->Enabled = true;
	edInterval->Enabled = true;
	btnRecorderStop->Enabled = false;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::btnRecorderClearClick(TObject *Sender)
{
	ScopedLock<Mutex> lock(mutexDeviceData);
	for (unsigned int i=0; i<current.GetChannelCount(); i++)
	{
		Data::Channel& channel = current.getChannel(i);
		channel.dataRec.clear();
	}
	bDataUpdate = true;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::tmrRecorderTimer(TObject *Sender)
{
	assert(display->GetMode() == Display::MODE_RECORDER);
	// execute manual trigger
	if (device)
		device->ManualTrigger();
	/** \todo check if previous sample already received */
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ActionExportPlotBitmapExecute(TObject *Sender)
{
	SaveDialog->DefaultExt = "bmp";
	SaveDialog->Filter = _("Bitmap (*.bmp)|*.bmp");
	SaveDialog->FilterIndex = 1;
	if (DirectoryExists(appSettings.asPlotPictureExportDir))
		SaveDialog->InitialDir = appSettings.asPlotPictureExportDir;
	SaveDialog->FileName = SaveDialog->InitialDir +
		((SaveDialog->InitialDir!="")?("\\"):("")) + FormatDateTime("yyyymmdd_hhnnss", Now());
	if (SaveDialog->Execute())
	{
		// refresh display (save dialog may be still visible on low spec PC)
		display->Update(Display::UPDATE_ALL);
		std::auto_ptr<Graphics::TBitmap> bitmap = CreatePlotBitmap();
		bitmap->PixelFormat = pf8bit;
		bitmap->SaveToFile(SaveDialog->FileName);
		appSettings.asPlotPictureExportDir = ExtractFileDir(SaveDialog->FileName);
	}
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ActionCalibrateSensitivityExecute(TObject *Sender)
{
	if (!frmCalibrateSensitivity->Visible)
		frmCalibrateSensitivity->Show();
	frmCalibrateSensitivity->BringToFront();
}
//---------------------------------------------------------------------------

int TfrmMain::CallbackCalculateCalibrationFn(float fVoltageHigh, int* const offset, float* const gain)
{
	float fZero, fCal, dummy;
	int status = display->GetCursorPhysical(Display::CURSOR2, dummy, fZero);
	status |= display->GetCursorPhysical(Display::CURSOR1, dummy, fCal);
	if (status)
		return status;
	// calculate offset
	int bitsZero =
		fZero/(device->capabilities->fSensitivityLevels[trbarSensitivity->Position]*current.getChannel(CurrentChannel()).calibration.gain) +
		current.getChannel(CurrentChannel()).calibration.offset;
	*offset = 0.5 + bitsZero;

	// calculate gain
	int bitsCal = fCal/(device->capabilities->fSensitivityLevels[trbarSensitivity->Position]*current.getChannel(CurrentChannel()).calibration.gain);
	bitsCal -= *offset;
	*gain = fVoltageHigh /
		( bitsCal * device->capabilities->fSensitivityLevels[trbarSensitivity->Position] );
	return 0;
}

int TfrmMain::CallbackStoreCalibrationFn(const int offset, const float gain)
{
	int channel = tabOscControlTop->TabIndex;
	int range = trbarSensitivity->Position;
	calibration[channel][range].offset = offset;
	calibration[channel][range].gain = gain;
	device->StoreCalibration(channel, range, offset, gain);
	ApplyCalibration();
	return 0;
}

void TfrmMain::CallbackPlotToolbarClickFn(int tag)
{
	// map tag to button
	TToolButton *Button;
	switch (tag)
	{
	case 1:
		Button = btnMouseCursor;
		break;
	case 2:
		Button = btnMouseMove;
		break;
	case 3:
		Button = btnMouseZoomXY;
		break;
	case 4:
		Button = btnMouseZoomX;
		break;
	case 5:
		Button = btnMouseZoomY;
		break;
	case 6:
		Button = btnZoomClear;
		break;
	default:
		return;
	}
	Button->Action->Execute();
	if (Button->Grouped)
		Button->Down = true;
}

void TfrmMain::ApplyCalibration(void)
{
	UpdateSensitivity();
}

void __fastcall TfrmMain::ActionExportDataExecute(TObject *Sender)
{
	SaveDialog->DefaultExt = "json";
	SaveDialog->Filter = _("JSON files (*.json)|*.json|CSV files (*.csv)|*.csv");
	SaveDialog->FilterIndex = 1;
	if (DirectoryExists(appSettings.asPlotDataExportDir))
		SaveDialog->InitialDir = appSettings.asPlotDataExportDir;
	SaveDialog->FileName = SaveDialog->InitialDir +
		((SaveDialog->InitialDir!="")?("\\"):("")) + FormatDateTime("yyyymmdd_hhnnss", Now());
	if (SaveDialog->Execute())
	{
		enum Recorder::E_FMT format;
		if (SaveDialog->FilterIndex == 1)
			format = Recorder::FMT_JSON;
		else
			format = Recorder::FMT_CSV;
		int status = Recorder::Store(SaveDialog->FileName.c_str(), format, current);
		appSettings.asPlotDataExportDir = ExtractFileDir(SaveDialog->FileName);

		AnsiString msg = _("Failed saving file");
		if (status)
			MessageBox(this->Handle, msg.c_str(),
				this->Caption.c_str(), MB_ICONEXCLAMATION);
	}
}
//---------------------------------------------------------------------------


void __fastcall TfrmMain::FormShow(TObject *Sender)
{
	static bool once = false;
	if (!once)
	{
		once = true;
		// frmLog is already created - set callback to preview log entries
		CLog::Instance()->callbackLog = frmLog->OnLog;
		frmLog->callbackClose = CallbackCloseLogFn;

		frmCalibrateSensitivity->callbackCalculate = CallbackCalculateCalibrationFn;
		frmCalibrateSensitivity->callbackStore = CallbackStoreCalibrationFn;

		UpdateDisplayVisibility();
        this->ActionViewPlotGridX->Checked = appSettings.GraphVisibility.bGridX;
       	this->ActionViewPlotGridY->Checked = appSettings.GraphVisibility.bGridY;
		this->ActionViewPlotGridLabelsX->Checked = appSettings.GraphVisibility.bGridLabelsX;
		this->ActionViewPlotGridLabelsY->Checked = appSettings.GraphVisibility.bGridLabelsY;
		this->ActionViewPlotCursors->Checked = appSettings.GraphVisibility.bCursor;
		this->ActionViewPlotGndMarkers->Checked = appSettings.GraphVisibility.bGndMarker;
		this->ActionViewPlotTriggerPosMarker->Checked = appSettings.GraphVisibility.bTriggerPosMarker;
		this->ActionViewPlotTriggerValMarker->Checked = appSettings.GraphVisibility.bTriggerValMarker;

		frmPlotToolbar->SetTransparecy(appSettings.Graph.ToolboxTransparency);

		// load device dlls
		asDevDllDir = ExtractFileDir(Application->ExeName) + "\\device";
		DeviceInterface::EnumerateDlls(asDevDllDir);

		Timer->Interval = 1000/appSettings.iDisplayRefreshFreq;
		display->ChangeSize(pnlCenter->Width, pnlCenter->Height);

		if (appSettings.asInputDeviceDllName != "")
		{
			AnsiString dllname = asDevDllDir + "\\" + appSettings.asInputDeviceDllName;
			for (unsigned int i=0; i<DeviceInterface::dlls.size(); i++)
			{
				if (DeviceInterface::dlls[i].name == dllname)
				{
					LoadDevice(dllname);
					break;
				}
			}
		}

		if (!device)
		{
			FirstRunDialog();
		}
		if (!device)
		{
			AnsiString msg = _("No input device dll is selected.\n"
				"Selecting input plugin is REQUIRED.\n"
				"Note: device plugins should be placed in \\device subdirectory of application.");
			AnsiString title = _("Execution aborted");
			MessageBox(this->Handle, msg.c_str(), title.c_str(), MB_ICONSTOP);
			Close();
			return;
		}
		SetMode(Display::MODE_OSCILLOSCOPE);
		frmPlotToolbar->callbackClick = CallbackPlotToolbarClickFn;

		frmLog->SetLogLinesLimit(appSettings.iMaxUiLogLines);

		frmStatistics->Left = appSettings.Statistics.iPosX;
		frmStatistics->Top = appSettings.Statistics.iPosY;
		if (appSettings.Statistics.bVisible)
			frmStatistics->Show();		
	}	
}
//---------------------------------------------------------------------------

/** \brief Set diaplay Y axis zoom & offset to fit trace into
*/
void __fastcall TfrmMain::ActionAutoFitYExecute(TObject *Sender)
{
    int channel_id = CurrentChannel();
	std::vector<float> *samples;
	Data::Channel& channel = current.getChannel(channel_id);
	if (display->GetMode() == Display::MODE_OSCILLOSCOPE)
		samples = &channel.dataOsc;
	else if (display->GetMode() == Display::MODE_SPECTRUM)
		samples = &channel.dataFFT;
	else if (display->GetMode() == Display::MODE_RECORDER)
		samples = &channel.dataRec;
	else
	{
		ShowMessage("unimplemented");
		return;
	}
	if (!samples->size())
		return;
	float fMin = std::numeric_limits<float>::max();
	float fMax = (-1.0) * fMin;

	std::vector<float>::const_iterator it;
	for (it = samples->begin(); it != samples->end(); ++it)
	{
		float val = *it;
		if (val < fMin)
			fMin = val;
		if (val > fMax)
			fMax = val;
	}

	if (fMin == fMax)
		return;

	int range = trbarSensitivity->Position;
	float fRange = (float)pow(2, device->capabilities->iBitsPerSample) *
		device->capabilities->fSensitivityLevels[range] *
		calibration[channel_id][range].gain;

	float fZoomX, fZoomY;
	display->GetZoom(fZoomX, fZoomY);
	display->SetZoom(Display::OFFSET_LEAVE, Display::OFFSET_LEAVE,
		fZoomX, fRange/(fMax-fMin));

	display->SetOffsetY((fMax + fMin)/2.0);
	UpdateZoomDisplay();
	bPaintUpdate = true;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ActionAutoFitXExecute(TObject *Sender)
{
	std::vector<float> *samples;
	Data* data = display->GetData();
	if (display->GetMode() == Display::MODE_OSCILLOSCOPE)
		samples = &data->getChannel(0).dataOsc;
	else if (display->GetMode() == Display::MODE_SPECTRUM)
		samples = &data->getChannel(0).dataFFT;
	else if (display->GetMode() == Display::MODE_RECORDER)
		samples = &data->getChannel(0).dataRec;
	else
	{
		ShowMessage("unimplemented");
		return;
	}
	if (!samples->size())
		return;

	float fZoomX, fZoomY;
	display->GetZoom(fZoomX, fZoomY);
	fZoomX = (float)pnlCenter->Width / samples->size();

	display->SetZoom(Display::OFFSET_CLEAR, Display::OFFSET_LEAVE, fZoomX, fZoomY);
	UpdateZoomDisplay();
	bPaintUpdate = true;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ActionLoadJsonFileExecute(TObject *Sender)
{
	OpenDialog->DefaultExt = "json";
	OpenDialog->Filter = _("JSON files (*.json)|*.json|All files (*.*)|*.*");
	OpenDialog->FilterIndex = 1;
	if (DirectoryExists(appSettings.asPlotDataExportDir))
		OpenDialog->InitialDir = appSettings.asPlotDataExportDir;
	if (OpenDialog->Execute())
	{
		int status = frmJsonViewer->LoadFile(OpenDialog->FileName);
		if (status != 0)
		{
			AnsiString msg = "Error loading JSON file";
			MessageBox(this->Handle, msg.c_str(),
				this->Caption.c_str(), MB_ICONEXCLAMATION);
			return;
		}
		bJsonViewerMode = true;
		ActionExportData->Enabled = false;
		UseData(frmJsonViewer->jsonData.data);
		display->Update(Display::UPDATE_ALL);
		pnlOscControlBottom->Visible = false;
		frmJsonViewer->Parent = pnlOscControl;
		frmJsonViewer->Visible = true;
		frmJsonViewer->callbackClose = CallbackJsonViewerClose;
		tabOscControlTop->Height = 22;
		tabOscControlTop->OnChange(NULL);
	}
}
//---------------------------------------------------------------------------

void TfrmMain::CallbackJsonViewerClose(void)
{
	bJsonViewerMode = false;
	ActionExportData->Enabled = true;
	pnlOscControlBottom->Visible = true;
	UseData(current);

	tabOscControlTop->Height = 109;
	tabOscControlTop->OnChange(NULL);
}

void TfrmMain::CallbackMs3ViewerClose(void)
{
	bMs3ViewerMode = false;
	ActionMs3Record->Enabled = true;
	ActionExportData->Enabled = true;
	pnlOscControlBottom->Visible = true;
	UseData(current);

	tabOscControlTop->Height = 109;
	tabOscControlTop->OnChange(NULL);
}

void TfrmMain::CallbackMs3ViewerDataUpdate(Data &ms3data)
{
	for (unsigned int i=0; i<ms3data.GetChannelCount(); i++) {
		filterChain.Process(ms3data.getChannel(i).dataOsc);
	}
	if (frmStatistics->Visible)
		frmStatistics->Process(ms3data, CurrentChannel());
	UseData(ms3data);
}

void TfrmMain::UseData(Data &data)
{
	unsigned int channel_cnt = data.GetChannelCount();
	if (tabOscControlTop->Tabs->Count != static_cast<int>(channel_cnt))
	{
	display->SetNumberOfChannels(channel_cnt);
	display->SetChannelColors(appSettings.GraphColors.traces);
	tabOscControlTop->Tabs->Clear();
	cbTriggerSource->Items->Clear();
	for (unsigned int i=0; i<channel_cnt; i++)
	{
		tabOscControlTop->Tabs->Add(GetChannelName(i));
		cbTriggerSource->Items->Add(GetChannelName(i));
	}
	tabOscControlTop->TabIndex = 0;
	}

	if (display->GetMode() == Display::MODE_SPECTRUM)
	{
		CalculateSpectrum();
	}
	display->SetData(&data);
}

void __fastcall TfrmMain::ActionShowDataFilterChainExecute(TObject *Sender)
{
	frmFilterSelector->ShowModal();
	if (frmFilterSelector->ModalResult == mrOk)
	{
		filterChain.Configure(frmFilterSelector->Get());
		bDataUpdate = true;
	}
}
//---------------------------------------------------------------------------

void TfrmMain::ReportError(int status, AnsiString description)
{
	if (status == 0)
		return;
	AnsiString asMsg = description + _("\nError code: ") + status;
	LOG(E_LOG_TRACE, "%s", asMsg.c_str());
	MessageBox(this->Handle, asMsg.c_str(), this->Caption.c_str(), MB_ICONEXCLAMATION);
}

void __fastcall TfrmMain::ActionMs3RecordExecute(TObject *Sender)
{
	if (ActionMs3Record->Checked)
	{
		SaveDialog->DefaultExt = "ms3";
		SaveDialog->Filter = _("MS3 files (*.ms3)|*.ms3");
		SaveDialog->FilterIndex = 1;
		if (DirectoryExists(appSettings.asPlotDataExportDir))
			SaveDialog->InitialDir = appSettings.asPlotDataExportDir;
		SaveDialog->FileName = SaveDialog->InitialDir +
			((SaveDialog->InitialDir!="")?("\\"):("")) + FormatDateTime("yyyymmdd_hhnnss", Now());
		if (SaveDialog->Execute())
		{
			if (FileExists(SaveDialog->FileName))
			{
				AnsiString msg = _("File already exists, file will be overwritten.") +
					"\r\n" + _("Are you sure you want to continue?");
				if (MessageBox(this->Handle, msg.c_str(),
					this->Caption.c_str(), MB_YESNO | MB_DEFBUTTON2 | MB_ICONEXCLAMATION) != IDYES)
				{
					ActionMs3Record->Checked = false;
					return;
				}
			}
			appSettings.asPlotDataExportDir = ExtractFileDir(SaveDialog->FileName);
			int status = Ms3Recorder::Instance().SetMode(Ms3Recorder::MODE_WRITE, SaveDialog->FileName);
			ReportError(status, _("Failed to open MS3 file for writing"));
			if (status)
				ActionMs3Record->Checked = false;
			else
				LOG(E_LOG_TRACE, "Recording data to %s file started", SaveDialog->FileName.c_str());
		}
		else
		{
			ActionMs3Record->Checked = false;
		}
	}
	else
	{
		int status = Ms3Recorder::Instance().SetMode(Ms3Recorder::MODE_IDLE);
		ReportError(status, _("Failed to close MS3 recorder/player."));
	}
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::ActionMs3LoadExecute(TObject *Sender)
{
	if (ActionMs3Record->Checked)
		ActionMs3Record->Execute();

	OpenDialog->DefaultExt = "ms3";
	OpenDialog->Filter = _("MS3 files (*.ms3)|*.ms3|All files (*.*)|*.*");
	OpenDialog->FilterIndex = 1;
	if (DirectoryExists(appSettings.asPlotDataExportDir))
		OpenDialog->InitialDir = appSettings.asPlotDataExportDir;
	if (OpenDialog->Execute())
	{
		int status = frmMs3Viewer->LoadFile(OpenDialog->FileName, CallbackMs3ViewerDataUpdate);
		if (status != 0)
		{
			AnsiString msg = _("Error loading MS3 file");
			MessageBox(this->Handle, msg.c_str(),
				this->Caption.c_str(), MB_ICONEXCLAMATION);
			bMs3ViewerMode = false;
			ActionMs3Record->Enabled = true;
			return;
		}
		bMs3ViewerMode = true;
		ActionMs3Record->Enabled = false;

		pnlOscControlBottom->Visible = false;
		frmMs3Viewer->Parent = pnlOscControl;
		frmMs3Viewer->Visible = true;
		tabOscControlTop->Height = 22;
		tabOscControlTop->OnChange(NULL);
		frmMs3Viewer->callbackClose = CallbackMs3ViewerClose;
	}
}
//---------------------------------------------------------------------------


void __fastcall TfrmMain::ActionMs3FileRepairExecute(TObject *Sender)
{
	OpenDialog->DefaultExt = "ms3";
	OpenDialog->Filter = _("MS3 files (*.ms3)|*.ms3|All files (*.*)|*.*");
	OpenDialog->FilterIndex = 1;
	if (DirectoryExists(appSettings.asPlotDataExportDir))
		OpenDialog->InitialDir = appSettings.asPlotDataExportDir;
	if (OpenDialog->Execute())
	{
		enum TfrmMs3FileRepair::E_STATUS status;
		frmMs3FileRepair->Execute(OpenDialog->FileName, status);
		MessageBox(this->Handle, frmMs3FileRepair->GetStatusDescription(status).c_str(),
			this->Caption.c_str(), MB_ICONINFORMATION);
	}
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::edIntervalExit(TObject *Sender)
{
	float interval = StrToFloatDef(edInterval->Text, -1.0);
	if (interval < Settings::fRecorderMinInterval || interval > Settings::fRecorderMaxInterval)
	{
		AnsiString msg = _("Recorder interval outside allowed range!");
		MessageBox(this->Handle, msg.c_str(), this->Caption.c_str(),
			MB_ICONEXCLAMATION);
	}
	else
	{
		appSettings.fRecorderInterval = interval;
	}
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::tabOscControlTopChange(TObject *Sender)
{
	int channel = tabOscControlTop->TabIndex;
	display->SetCurrentChannel(channel);
	UpdateZoomDisplay();
	bPaintUpdate = true;
	if (bJsonViewerMode == false)
	{
		trbarSensitivity->Position = device->settings.pSensitivity[channel];
		int range = trbarSensitivity->Position;
		float fDivVal = (float)pow(2, device->capabilities->iBitsPerSample) / (float)10 *
			device->capabilities->fSensitivityLevels[range] *
			calibration[channel][range].gain;
		lblSensitivityVal->Caption =
			ValueStringFormatter::VoltageToStr(fDivVal, 1).c_str();
		lblSensitivityVal->Caption = lblSensitivityVal->Caption + "/div";
		TNotifyEvent callback;
		callback = cbCouplingType->OnChange;
		cbCouplingType->OnChange = NULL;
		for (unsigned int i=0; i<device->capabilities->iCouplingTypeCount; i++)
		{
			if (device->capabilities->iCouplingType[i] == device->settings.eCouplingType[channel])
			{
				cbCouplingType->ItemIndex = i;
				break;
			}
		}
		cbCouplingType->OnChange = callback;
	}
	else
	{
    	frmJsonViewer->SetCurrentChannel(channel);
	}

	if (display->GetMode() == Display::MODE_SPECTRUM)
	{
		CalculateSpectrum();
	}

	if (frmStatistics->Visible)
	{
		frmStatistics->Process(current, CurrentChannel());
    }
}
//---------------------------------------------------------------------------

int TfrmMain::CurrentChannel(void)
{
	return tabOscControlTop->TabIndex;
}

void __fastcall TfrmMain::tabOscControlTopDrawTab(TCustomTabControl *Control,
	  int TabIndex, const TRect &Rect, bool Active)
{
	TCanvas *pCanvas = Control->Canvas;
	if (display)
	{
		RECT rect = static_cast<RECT>(Rect);
		rect.top = Rect.Bottom - 4;
		FillRect(pCanvas->Handle, &rect, CreateSolidBrush(display->GetChannelColor(TabIndex)));
	}
	pCanvas->Brush->Color = clBtnFace;
	RECT R = static_cast<RECT>(Rect);
	AnsiString text(tabOscControlTop->Tabs->Strings[TabIndex]);
	DrawText(pCanvas->Handle, text.c_str(), text.Length(), &R,
			 DT_CENTER | DT_TOP | DT_SINGLELINE);
}
//---------------------------------------------------------------------------

void TfrmMain::UpdateDisplayVisibility(void)
{
	display->SetVisibility(Display::GRID_X, appSettings.GraphVisibility.bGridX);
	display->SetVisibility(Display::GRID_Y, appSettings.GraphVisibility.bGridY);
	display->SetVisibility(Display::GRID_LABELS_X, appSettings.GraphVisibility.bGridLabelsX);
	display->SetVisibility(Display::GRID_LABELS_Y, appSettings.GraphVisibility.bGridLabelsY);
	display->SetVisibility(Display::CURSOR, appSettings.GraphVisibility.bCursor);
	display->SetVisibility(Display::GND_MARKER, appSettings.GraphVisibility.bGndMarker);
	if ((upcoming.trigger_type == TRIGGER_SLOPE_UP || upcoming.trigger_type == TRIGGER_SLOPE_DOWN) &&
		device->hasTriggerPositionControl())
	{
		display->SetVisibility(Display::TRIGGER_POS_MARKER, appSettings.GraphVisibility.bTriggerPosMarker);
	}
	else
	{
		display->SetVisibility(Display::TRIGGER_POS_MARKER, false);
	}
	if (upcoming.trigger_type == TRIGGER_SLOPE_UP || upcoming.trigger_type == TRIGGER_SLOPE_DOWN)
	{
		display->SetVisibility(Display::TRIGGER_VAL_MARKER, appSettings.GraphVisibility.bTriggerValMarker);
	}
	else
	{
		display->SetVisibility(Display::TRIGGER_VAL_MARKER, false);
	}
}

void __fastcall TfrmMain::ActionViewPlotExecute(TObject *Sender)
{
	// plot elements visibility changed
	appSettings.GraphVisibility.bGridX = this->ActionViewPlotGridX->Checked;
	appSettings.GraphVisibility.bGridY = this->ActionViewPlotGridY->Checked;
	appSettings.GraphVisibility.bGridLabelsX = this->ActionViewPlotGridLabelsX->Checked;
	appSettings.GraphVisibility.bGridLabelsY = this->ActionViewPlotGridLabelsY->Checked;
	appSettings.GraphVisibility.bCursor = this->ActionViewPlotCursors->Checked;
	appSettings.GraphVisibility.bGndMarker = this->ActionViewPlotGndMarkers->Checked;
	appSettings.GraphVisibility.bTriggerPosMarker = this->ActionViewPlotTriggerPosMarker->Checked;	
	appSettings.GraphVisibility.bTriggerValMarker = this->ActionViewPlotTriggerValMarker->Checked;	

	UpdateDisplayVisibility();
	bPaintUpdate = true;	
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::FormMouseWheelDown(TObject *Sender, TShiftState Shift,
	  TPoint &MousePos, bool &Handled)
{
	TPoint P = PaintBoxCenter->ScreenToClient(Mouse->CursorPos);
	if (P.x >= 0 && P.y >= 0 && P.x < PaintBoxCenter->Width && P.y < PaintBoxCenter->Height)
	{
		Handled = true;
		float fZoomX, fZoomY;
		display->GetZoom(fZoomX, fZoomY);
		float fDivY = display->GetDivY();
		if (Shift.Contains(ssShift))
		{
			fZoomY /= fZoomChangeGrid;
		}
		else
		{
			fZoomY = fZoomY * fDivY/GetNextRoundHigherValue(fDivY);
		}
		display->SetZoom(Display::OFFSET_LEAVE, Display::OFFSET_KEEP_GND_POSITION,
			fZoomX, fZoomY);
		UpdateZoomDisplay();

		bPaintUpdate = true;
		return;
	}
	Handled = false;
}
//---------------------------------------------------------------------------

void __fastcall TfrmMain::FormMouseWheelUp(TObject *Sender, TShiftState Shift,
	  TPoint &MousePos, bool &Handled)
{
	TPoint P = PaintBoxCenter->ScreenToClient(Mouse->CursorPos);
	if (P.x >= 0 && P.y >= 0 && P.x < PaintBoxCenter->Width && P.y < PaintBoxCenter->Height)
	{
		Handled = true;
		Handled = true;
		float fZoomX, fZoomY;
		display->GetZoom(fZoomX, fZoomY);
		float fDivY = display->GetDivY();
		if (Shift.Contains(ssShift))
		{
			fZoomY *= fZoomChangeGrid;
		}
		else
		{
			fZoomY = fZoomY * fDivY/GetNextRoundLowerValue(fDivY);
		}
		display->SetZoom(Display::OFFSET_LEAVE, Display::OFFSET_KEEP_GND_POSITION,
			fZoomX, fZoomY);
		UpdateZoomDisplay();

		bPaintUpdate = true;
		return;
	}
	Handled = false;
	
}
//---------------------------------------------------------------------------

